from __future__ import annotations

from functools import partial
from typing import Any, Callable, Dict, TypeVar, Union

import astropy.units as u
import numpy as np

from zodipy._bandpass import Bandpass
from zodipy._constants import SPECIFIC_INTENSITY_UNITS
from zodipy._ipd_comps import ComponentLabel
from zodipy._ipd_model import RRM, InterplanetaryDustModel, Kelsall

InterplanetaryDustModelT = TypeVar("InterplanetaryDustModelT", bound=InterplanetaryDustModel)

"""Return the source parameters for a given bandpass and model.
Must match arguments in the emission fns."""
GetSourceParametersFn = Callable[
    [Bandpass, InterplanetaryDustModelT, Callable],
    Dict[Union[ComponentLabel, str], Any],
]


def get_source_parameters_kelsall_comp(
    bandpass: Bandpass, model: Kelsall, interpolator: Callable
) -> dict[ComponentLabel | str, dict[str, Any]]:
    if not bandpass.frequencies.unit.is_equivalent(model.spectrum.unit):
        bandpass.switch_convention()

    spectrum = (
        model.spectrum.to_value(u.Hz)
        if model.spectrum.unit.is_equivalent(u.Hz)
        else model.spectrum.to_value(u.micron)
    )

    interpolator = partial(interpolator, x=spectrum)

    source_parameters: dict[ComponentLabel | str, dict[str, Any]] = {}
    for comp_label in model.comps:
        source_parameters[comp_label] = {}
        emissivity = interpolator(y=model.emissivities[comp_label])(bandpass.frequencies.value)
        if model.albedos is not None:
            albedo = interpolator(y=model.albedos[comp_label])(bandpass.frequencies.value)
        else:
            albedo = 0

        if bandpass.frequencies.size > 1:
            emissivity = bandpass.integrate(emissivity)
            albedo = bandpass.integrate(albedo)

        source_parameters[comp_label]["emissivity"] = emissivity
        source_parameters[comp_label]["albedo"] = albedo

    if model.phase_coefficients is not None:
        phase_coefficients = interpolator(y=np.asarray(model.phase_coefficients))(
            bandpass.frequencies.value
        )
        phase_coefficients = interpolator(y=np.asarray(model.phase_coefficients))(
            bandpass.frequencies.value
        )
    else:
        phase_coefficients = np.repeat(np.zeros((3, 1)), repeats=bandpass.frequencies.size, axis=-1)

    if model.solar_irradiance is not None:
        solar_irradiance = interpolator(y=model.solar_irradiance)(bandpass.frequencies.value)
        solar_irradiance = u.Quantity(solar_irradiance, "MJy /sr").to_value(
            SPECIFIC_INTENSITY_UNITS, equivalencies=u.spectral()
        )
    else:
        solar_irradiance = 0

    if bandpass.frequencies.size > 1:
        phase_coefficients = bandpass.integrate(phase_coefficients)
        solar_irradiance = bandpass.integrate(solar_irradiance)
    source_parameters["common"] = {}
    source_parameters["common"]["phase_coefficients"] = tuple(phase_coefficients)
    source_parameters["common"]["solar_irradiance"] = solar_irradiance
    source_parameters["common"]["T_0"] = model.T_0
    source_parameters["common"]["delta"] = model.delta

    return source_parameters


def get_source_parameters_rmm(
    bandpass: Bandpass, model: RRM, interpolator: Callable
) -> dict[ComponentLabel | str, dict[str, Any]]:
    if not bandpass.frequencies.unit.is_equivalent(model.spectrum.unit):
        bandpass.switch_convention()

    spectrum = (
        model.spectrum.to_value(u.Hz)
        if model.spectrum.unit.is_equivalent(u.Hz)
        else model.spectrum.to_value(u.micron)
    )

    source_parameters: dict[ComponentLabel | str, dict[str, Any]] = {}
    calibration = interpolator(x=spectrum, y=model.calibration)(bandpass.frequencies.value)
    calibration = u.Quantity(calibration, u.MJy / u.AU).to_value(u.Jy / u.cm)

    if bandpass.frequencies.size > 1:
        calibration = bandpass.integrate(calibration)

    for comp_label in model.comps:
        source_parameters[comp_label] = {}
        source_parameters[comp_label]["T_0"] = model.T_0[comp_label]
        source_parameters[comp_label]["delta"] = model.delta[comp_label]

    source_parameters["common"] = {"calibration": calibration}

    return source_parameters


SOURCE_PARAMS_MAPPING: dict[type[InterplanetaryDustModel], GetSourceParametersFn] = {
    Kelsall: get_source_parameters_kelsall_comp,
    RRM: get_source_parameters_rmm,
}



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

import astropy.units as u
import numpy as np

from zodipy._ipd_dens_funcs import construct_density_partials
from zodipy.model_registry import model_registry

if TYPE_CHECKING:
    import numpy.typing as npt

DEFAULT_EARTH_POS = u.Quantity([1, 0, 0], u.AU)


def tabulate_density(
    grid: npt.NDArray[np.floating] | Sequence[npt.NDArray[np.floating]],
    model: str = "DIRBE",
    earth_pos: u.Quantity[u.AU] = DEFAULT_EARTH_POS,
) -> npt.NDArray[np.float64]:
    """Return the tabulated densities of the zodiacal components for a given grid.

    Args:
        grid: A cartesian mesh grid (x, y, z).
        model: Name of interplanetary dust model supported by ZodiPy.
        earth_pos: Position of the Earth in AU.

    Returns:
        density_grid: The tabulated zodiacal component densities.

    """
    ipd_model = model_registry.get_model(model)

    if not isinstance(grid, np.ndarray):
        grid = np.asarray(grid)

    # Prepare attributes and variables for broadcasting with the grid
    earth_position = np.reshape(earth_pos.to(u.AU).value, (3, 1, 1, 1))
    for comp in ipd_model.comps.values():
        comp.X_0 = np.reshape(comp.X_0, (3, 1, 1, 1))

    partials = construct_density_partials(
        list(ipd_model.comps.values()), {"X_earth": earth_position}
    )

    density_grid = np.zeros((len(ipd_model.comps), *grid.shape[1:]))
    for idx, partial in enumerate(partials):
        density_grid[idx] = partial(grid)

    # Revert broadcasting reshapes
    for comp in ipd_model.comps.values():
        comp.X_0 = np.reshape(comp.X_0, (3, 1))

    return density_grid



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Mapping, Sequence

if TYPE_CHECKING:
    from zodipy._ipd_comps import Component, ComponentLabel
    from zodipy._types import FrequencyOrWavelength


@dataclass
class InterplanetaryDustModel(ABC):
    """Base class for interplanetary dust models.

    Args:
        comps: Mapping of component labels to component classes.
        spectrum: Frequency or wavelength over which the model is valid.

    """

    comps: Mapping[ComponentLabel, Component]
    spectrum: FrequencyOrWavelength

    def to_dict(self) -> dict:
        """Return a dictionary representation of the model."""
        _dict: dict = {}
        for key, value in vars(self).items():
            if key == "comps":
                _dict[key] = {}
                for comp_key, comp_value in value.items():
                    _dict[key][comp_key.value] = {
                        k: v
                        for k, v in vars(comp_value).items()
                        if comp_value.__dataclass_fields__[k].init
                    }
            elif isinstance(value, dict):
                _dict[key] = {k.value: v for k, v in value.items()}
            else:
                _dict[key] = value

        return _dict


@dataclass
class Kelsall(InterplanetaryDustModel):
    """Kelsall et al. (1998) model."""

    T_0: float
    delta: float
    emissivities: Mapping[ComponentLabel, Sequence[float]]
    albedos: Mapping[ComponentLabel, Sequence[float]] | None = None
    solar_irradiance: Sequence[float] | None = None  # In units of MJy/sr
    phase_coefficients: Sequence[Sequence[float]] | None = None


@dataclass
class RRM(InterplanetaryDustModel):
    """Rowan-Robinson and May (2013) model."""

    T_0: Mapping[ComponentLabel, float]
    delta: Mapping[ComponentLabel, float]
    calibration: Sequence[float]


@dataclass
class InterplanetaryDustModelRegistry:
    """Container for registered models."""

    _registry: dict[str, InterplanetaryDustModel] = field(init=False, default_factory=dict)

    @property
    def models(self) -> list[str]:
        return list(self._registry.keys())

    def register_model(
        self,
        name: str,
        model: InterplanetaryDustModel,
    ) -> None:
        if (name := name.lower()) in self._registry:
            msg = f"a model by the name {name!s} is already registered."
            raise ValueError(msg)

        self._registry[name] = model

    def get_model(self, name: str) -> InterplanetaryDustModel:
        if (name := name.lower()) not in self._registry:
            msg = (
                f"{name!r} is not a registered Interplanetary Dust model. "
                f"Avaliable models are: {', '.join(self._registry)}."
            )
            raise ValueError(msg)

        return self._registry[name]


model_registry = InterplanetaryDustModelRegistry()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import numpy as np
import numpy.typing as npt

from zodipy._ipd_model import RRM, InterplanetaryDustModel, Kelsall
from zodipy._source_funcs import (
    get_dust_grain_temperature,
    get_phase_function,
    get_scattering_angle,
)

if TYPE_CHECKING:
    from zodipy._ipd_dens_funcs import ComponentDensityFn

"""
Function that return the zodiacal emission at a step along all lines of sight given
a zodiacal model.
"""
GetCompEmissionAtStepFn = Callable[..., npt.NDArray[np.float64]]


def kelsall(
    r: npt.NDArray[np.float64],
    start: np.float64,
    stop: npt.NDArray[np.float64],
    X_obs: npt.NDArray[np.float64],
    u_los: npt.NDArray[np.float64],
    get_density_function: ComponentDensityFn,
    T_0: float,
    delta: float,
    emissivity: np.float64,
    albedo: np.float64,
    phase_coefficients: tuple[float, ...],
    solar_irradiance: np.float64,
    bp_interpolation_table: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Kelsall uses common line of sight grid from obs to 5.2 AU."""
    # Convert the quadrature range from [-1, 1] to the true ecliptic positions
    R_los = ((stop - start) / 2) * r + (stop + start) / 2
    X_los = R_los * u_los
    X_helio = X_los + X_obs
    R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)

    temperature = get_dust_grain_temperature(R_helio, T_0, delta)
    blackbody_emission = np.interp(temperature, *bp_interpolation_table)
    emission = (1 - albedo) * (emissivity * blackbody_emission)

    if albedo != 0:
        solar_flux = solar_irradiance / R_helio**2
        scattering_angle = get_scattering_angle(R_los, R_helio, X_los, X_helio)
        phase_function = get_phase_function(scattering_angle, phase_coefficients)

        emission += albedo * solar_flux * phase_function

    return emission * get_density_function(X_helio)


def rrm(
    r: npt.NDArray[np.float64],
    start: npt.NDArray[np.float64],
    stop: npt.NDArray[np.float64],
    X_obs: npt.NDArray[np.float64],
    u_los: npt.NDArray[np.float64],
    get_density_function: ComponentDensityFn,
    T_0: float,
    delta: float,
    calibration: np.float64,
    bp_interpolation_table: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """RRM is implented with component specific line-of-sight grids."""
    # Convert the quadrature range from [-1, 1] to the true ecliptic positions
    R_los = ((stop - start) / 2) * r + (stop + start) / 2
    X_los = R_los * u_los
    X_helio = X_los + X_obs
    R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)

    temperature = get_dust_grain_temperature(R_helio, T_0, delta)
    blackbody_emission = np.interp(temperature, *bp_interpolation_table)

    return blackbody_emission * get_density_function(X_helio) * calibration


EMISSION_MAPPING: dict[type[InterplanetaryDustModel], GetCompEmissionAtStepFn] = {
    Kelsall: kelsall,
    RRM: rrm,
}



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np

from ._constants import c, h, k_B

if TYPE_CHECKING:
    import numpy.typing as npt


def get_blackbody_emission(
    freq: float | npt.NDArray[np.float64], T: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    """Return the blackbody emission given a sequence of frequencies and temperatures.

    Args:
        freq: Frequency [Hz].
        T: Temperature of the blackbody [K].

    Returns:
        Blackbody emission [W / m^2 Hz sr].

    """
    term1 = (2 * h * freq**3) / c**2
    term2 = np.expm1((h * freq) / (k_B * T))
    return term1 / term2


def get_dust_grain_temperature(
    R: npt.NDArray[np.float64], T_0: float, delta: float
) -> npt.NDArray[np.float64]:
    """Return the dust grain temperature given a radial distance from the Sun.

    Args:
        R: Radial distance from the sun in ecliptic heliocentric coordinates [AU / 1AU].
        T_0: Temperature of dust grains located 1 AU from the Sun [K].
        delta: Powerlaw index.

    Returns:
        Dust grain temperature [K].

    """
    return T_0 * R**-delta


def get_scattering_angle(
    R_los: float | npt.NDArray[np.float64],
    R_helio: npt.NDArray[np.float64],
    X_los: npt.NDArray[np.float64],
    X_helio: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Return the scattering angle between the Sun and a point along the line of sight.

    Args:
        R_los: Distance from observer to a point along the line of sight.
        R_helio: Distance from the Sun to a point along the line of sight.
        X_los: Vector pointing from the observer to a point along the line of sight.
        X_helio: Heliocentric position of point along the line of sight.

    Returns:
        Scattering angle.

    """
    cos_theta = (X_los * X_helio).sum(axis=0) / (R_los * R_helio)
    cos_theta = np.clip(cos_theta, -1, 1)
    return np.arccos(-cos_theta)


def get_phase_function(
    Theta: npt.NDArray[np.float64], C: tuple[float, ...]
) -> npt.NDArray[np.float64]:
    """Return the phase function.

    Args:
        Theta: Scattering angle.
        C: Phase function parameters.

    Returns:
        The Phase funciton.

    """
    phase_normalization = _get_phase_normalization(C)
    return phase_normalization * (C[0] + C[1] * Theta + np.exp(C[2] * Theta))


@lru_cache
def _get_phase_normalization(C: tuple[float, ...]) -> float:
    """Return the analyitcal integral for the phase normalization factor N."""
    int_term1 = 2 * np.pi
    int_term2 = 2 * C[0]
    int_term3 = np.pi * C[1]
    int_term4 = (np.exp(C[2] * np.pi) + 1) / (C[2] ** 2 + 1)
    return 1 / (int_term1 * (int_term2 + int_term3 + int_term4))



import contextlib

import pkg_resources

from zodipy import comps, source_params
from zodipy._contour import tabulate_density
from zodipy.model_registry import model_registry
from zodipy.zodipy import Zodipy

contextlib.suppress(pkg_resources.DistributionNotFound)

__all__ = (
    "Zodipy",
    "model_registry",
    "comps",
    "source_params",
    "tabulate_density",
)



from typing import Sequence, Tuple, Union

import astropy.units as u
import healpy as hp
import numpy as np
import numpy.typing as npt

from ._ipd_model import InterplanetaryDustModel
from ._types import FrequencyOrWavelength, Pixels, SkyAngles


def get_validated_freq(
    freq: FrequencyOrWavelength, model: InterplanetaryDustModel, extrapolate: bool
) -> FrequencyOrWavelength:
    """Validate user inputted frequency."""
    if not isinstance(freq, u.Quantity):
        msg = "Frequency must be an astropy Quantity."
        raise TypeError(msg)

    if freq.unit.is_equivalent(u.Hz):
        freq = freq.to(u.Hz)
    elif freq.unit.is_equivalent(u.micron):
        freq = freq.to(u.micron)
    else:
        msg = "Frequency must be in units compatible with Hz or micron."
        raise u.UnitsError(msg)

    if extrapolate:
        return freq

    freq_in_spectrum_units = freq.to(model.spectrum.unit, equivalencies=u.spectral())
    lower_freq_range = model.spectrum.min()
    upper_freq_range = model.spectrum.max()

    if freq_in_spectrum_units.isscalar:
        freq_is_in_range = lower_freq_range <= freq_in_spectrum_units <= upper_freq_range
    else:
        freq_is_in_range = all(
            lower_freq_range.value <= nu <= upper_freq_range.value and nu
            for nu in freq_in_spectrum_units.value
        )

    if not freq_is_in_range:
        msg = f"Model is only valid in the [{lower_freq_range}," f" {upper_freq_range}] range."
        raise ValueError(msg)

    return freq


def get_validated_and_normalized_weights(
    weights: Union[Sequence[float], npt.NDArray[np.floating], None],
    freq: FrequencyOrWavelength,
) -> npt.NDArray[np.float64]:
    """Validate user inputted weights."""
    if weights is None and freq.size > 1:
        msg = "Bandpass weights must be specified if more than one frequency is given."
        raise ValueError(msg)

    if weights is not None:
        if freq.size != len(weights):
            msg = "Number of frequencies and weights must be the same."
            raise ValueError(msg)
        if np.any(np.diff(freq) < 0):
            msg = "Bandpass frequencies must be strictly increasing."
            raise ValueError(msg)

        normalized_weights = np.asarray(weights, dtype=np.float64)
    else:
        normalized_weights = np.array([1], dtype=np.float64)

    if normalized_weights.size > 1:
        return normalized_weights / np.trapz(normalized_weights, freq.value)

    return normalized_weights


@u.quantity_input(theta=[u.deg, u.rad], phi=[u.deg, u.rad])
def get_validated_ang(
    theta: SkyAngles, phi: SkyAngles, lonlat: bool
) -> Tuple[SkyAngles, SkyAngles]:
    """Validate user inputted sky angles."""
    theta = theta.to(u.deg) if lonlat else theta.to(u.rad)
    phi = phi.to(u.deg) if lonlat else phi.to(u.rad)

    if theta.isscalar:
        theta = u.Quantity([theta])
    if phi.isscalar:
        phi = u.Quantity([phi])

    return theta, phi


def get_validated_pix(pixels: Pixels, nside: int) -> Pixels:
    """Validate user inputted pixels."""
    if (np.max(pixels) > hp.nside2npix(nside)) or (np.min(pixels) < 0):
        msg = "invalid pixel number given nside"
        raise ValueError(msg)

    if np.ndim(pixels) == 0:
        return np.expand_dims(pixels, axis=0)

    return pixels



from __future__ import annotations

import multiprocessing
import platform
from functools import partial
from typing import TYPE_CHECKING, Callable, Literal, Sequence

import astropy.units as u
import healpy as hp
import numpy as np
from astropy.coordinates import solar_system_ephemeris
from scipy.interpolate import interp1d

from zodipy._bandpass import get_bandpass_interpolation_table, validate_and_get_bandpass
from zodipy._constants import SPECIFIC_INTENSITY_UNITS
from zodipy._emission import EMISSION_MAPPING
from zodipy._interpolate_source import SOURCE_PARAMS_MAPPING
from zodipy._ipd_comps import ComponentLabel
from zodipy._ipd_dens_funcs import construct_density_partials_comps
from zodipy._line_of_sight import get_line_of_sight_start_and_stop_distances
from zodipy._sky_coords import get_obs_and_earth_positions
from zodipy._unit_vectors import get_unit_vectors_from_ang, get_unit_vectors_from_pixels
from zodipy._validators import get_validated_ang, get_validated_pix
from zodipy.model_registry import model_registry

if TYPE_CHECKING:
    import numpy.typing as npt
    from astropy.time import Time

    from zodipy._types import FrequencyOrWavelength, ParameterDict, Pixels, SkyAngles


PLATFORM = platform.system().lower()
SYS_PROC_START_METHOD = "fork" if "windows" not in PLATFORM else None


class Zodipy:
    """Interface for simulating zodiacal emission.

    This class provides methods for simulating zodiacal emission given observer pointing
    either in sky angles or through HEALPix pixels.

    Args:
        model (str): Name of the interplanetary dust model to use in the simulations.
            Defaults to DIRBE.
        gauss_quad_degree (int): Order of the Gaussian-Legendre quadrature used to evaluate
            the line-of-sight integral in the simulations. Default is 50 points.
        interp_kind (str): Interpolation kind used to interpolate relevant model parameters.
            Defaults to 'linear'. For more information on available interpolation methods,
            please visit the [Scipy documentation](
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html).
        extrapolate (bool): If `True` all spectral quantities in the selected model are
            extrapolated to the requested frequencies or wavelengths. If `False`, an
            exception is raised on requested frequencies/wavelengths outside of the
            valid model range. Default is `False`.
        ephemeris (str): Ephemeris used to compute the positions of the observer and the
            Earth. Defaults to 'de432s', which requires downloading (and caching) a ~10MB
            file. For more information on available ephemeridis, please visit the [Astropy
            documentation](https://docs.astropy.org/en/stable/coordinates/solarsystem.html)
        solar_cut (u.Quantity[u.deg]): Cutoff angle from the sun in degrees. The emission
            for all the pointing with angular distance between the sun smaller than
            `solar_cutoff` are masked. Defaults to `None`.
        solar_cut_fill_value (float): Fill value for pixels masked with `solar_cut`.
            Defaults to `np.nan`.
        parallel (bool): If `True`, input pointing will be split among several cores, and
            the emission will be computed in parallel. Default is `False`.
        n_proc (int): Number of cores to use when `parallel` is `True`. Defaults is `None`,
            which uses all available cores.

    """

    def __init__(
        self,
        model: str = "dirbe",
        gauss_quad_degree: int = 50,
        extrapolate: bool = False,
        interp_kind: str = "linear",
        ephemeris: str = "de432s",
        solar_cut: u.Quantity[u.deg] | None = None,
        solar_cut_fill_value: float = np.nan,
        parallel: bool = False,
        n_proc: int | None = None,
    ) -> None:
        self.model = model
        self.gauss_quad_degree = gauss_quad_degree
        self.extrapolate = extrapolate
        self.interp_kind = interp_kind
        self.ephemeris = ephemeris
        self.solar_cut = solar_cut.to(u.rad) if solar_cut is not None else solar_cut
        self.solar_cut_fill_value = solar_cut_fill_value
        self.parallel = parallel
        self.n_proc = n_proc

        self._interpolator = partial(
            interp1d,
            kind=self.interp_kind,
            fill_value="extrapolate" if self.extrapolate else np.nan,
        )
        self._ipd_model = model_registry.get_model(model)
        self._gauss_points_and_weights = np.polynomial.legendre.leggauss(gauss_quad_degree)

    @property
    def supported_observers(self) -> list[str]:
        """Return a list of available observers given an ephemeris."""
        return [*list(solar_system_ephemeris.bodies), "semb-l2"]

    def get_parameters(self) -> ParameterDict:
        """Return a dictionary containing the interplanetary dust model parameters."""
        return self._ipd_model.to_dict()

    def update_parameters(self, parameters: ParameterDict) -> None:
        """Update the interplanetary dust model parameters.

        Args:
            parameters: Dictionary of parameters to update. The keys must be the names
                of the parameters as defined in the model. To get the parameters dict
                of an existing model, use `Zodipy("dirbe").get_parameters()`.

        """
        _dict = parameters.copy()
        _dict["comps"] = {}
        for key, value in parameters.items():
            if key == "comps":
                for comp_key, comp_value in value.items():
                    _dict["comps"][ComponentLabel(comp_key)] = type(
                        self._ipd_model.comps[ComponentLabel(comp_key)]
                    )(**comp_value)
            elif isinstance(value, dict):
                _dict[key] = {ComponentLabel(k): v for k, v in value.items()}

        self._ipd_model = self._ipd_model.__class__(**_dict)

    def get_emission_ang(
        self,
        freq: FrequencyOrWavelength,
        theta: SkyAngles,
        phi: SkyAngles,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        weights: Sequence[float] | npt.NDArray[np.floating] | None = None,
        lonlat: bool = False,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Return the simulated zodiacal emission given angles on the sky.

        The pointing, for which to compute the emission, is specified in form of angles on
        the sky given by `theta` and `phi`.

        Args:
            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
                a bandpass over which to evaluate the zodiacal emission. The frequencies
                must be strictly increasing.
            theta: Angular co-latitude coordinate of a point, or a sequence of points, on
                the celestial sphere. Must be in the range [0, π] rad. Units must be either
                radians or degrees.
            phi: Angular longitude coordinate of a point, or a sequence of points, on the
                celestial sphere. Must be in the range [0, 2π] rad. Units must be either
                radians or degrees.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            weights: Bandpass weights corresponding the the frequencies in `freq`. The weights
                are assumed to be given in spectral radiance units (Jy/sr).
            lonlat: If True, input angles (`theta`, `phi`) are assumed to be longitude and
                latitude, otherwise, they are co-latitude and longitude.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """
        theta, phi = get_validated_ang(theta=theta, phi=phi, lonlat=lonlat)

        unique_angles, indicies = np.unique(np.asarray([theta, phi]), return_inverse=True, axis=1)
        unit_vectors = get_unit_vectors_from_ang(
            coord_in=coord_in,
            theta=unique_angles[0],
            phi=unique_angles[1],
            lonlat=lonlat,
        )

        return self._compute_emission(
            freq=freq,
            weights=weights,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=indicies,
            return_comps=return_comps,
        )

    def get_emission_pix(
        self,
        freq: FrequencyOrWavelength,
        pixels: Pixels,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        weights: Sequence[float] | npt.NDArray[np.floating] | None = None,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Return the simulated zodiacal emission given pixel numbers.

        The pixel numbers represent the pixel indicies on a HEALPix grid with resolution
        given by `nside`.

        Args:
            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
                a bandpass over which to evaluate the zodiacal emission. The frequencies
                must be strictly increasing.
            pixels: HEALPix pixel indicies representing points on the celestial sphere.
            nside: HEALPix resolution parameter of the pixels and the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            weights: Bandpass weights corresponding the the frequencies in `freq`. The weights
                are assumed to be given in spectral radiance units (Jy/sr).
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """
        pixels = get_validated_pix(pixels=pixels, nside=nside)

        unique_pixels, indicies = np.unique(pixels, return_inverse=True)
        unit_vectors = get_unit_vectors_from_pixels(
            coord_in=coord_in,
            pixels=unique_pixels,
            nside=nside,
        )

        return self._compute_emission(
            freq=freq,
            weights=weights,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=indicies,
            pixels=unique_pixels,
            nside=nside,
            return_comps=return_comps,
        )

    def get_binned_emission_ang(
        self,
        freq: FrequencyOrWavelength,
        theta: SkyAngles,
        phi: SkyAngles,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        weights: Sequence[float] | npt.NDArray[np.floating] | None = None,
        lonlat: bool = False,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Return the simulated binned zodiacal emission given angles on the sky.

        The pointing, for which to compute the emission, is specified in form of angles on
        the sky given by `theta` and `phi`. The emission is binned to a HEALPix map with
        resolution given by `nside`.

        Args:
            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
                a bandpass over which to evaluate the zodiacal emission. The frequencies
                must be strictly increasing.
            theta: Angular co-latitude coordinate of a point, or a sequence of points, on
                the celestial sphere. Must be in the range [0, π] rad. Units must be either
                radians or degrees.
            phi: Angular longitude coordinate of a point, or a sequence of points, on the
                celestial sphere. Must be in the range [0, 2π] rad. Units must be either
                radians or degrees.
            nside: HEALPix resolution parameter of the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            weights: Bandpass weights corresponding the the frequencies in `freq`. The weights
                are assumed to be given in spectral radiance units (Jy/sr).
            lonlat: If True, input angles `theta`, `phi` are assumed to be longitude and
                latitude, otherwise, they are co-latitude and longitude.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """
        theta, phi = get_validated_ang(theta=theta, phi=phi, lonlat=lonlat)

        unique_angles, counts = np.unique(np.asarray([theta, phi]), return_counts=True, axis=1)
        unique_pixels = hp.ang2pix(nside, *unique_angles, lonlat=lonlat)
        unit_vectors = get_unit_vectors_from_ang(
            coord_in=coord_in,
            theta=unique_angles[0],
            phi=unique_angles[1],
            lonlat=lonlat,
        )

        return self._compute_emission(
            freq=freq,
            weights=weights,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=counts,
            binned=True,
            pixels=unique_pixels,
            nside=nside,
            return_comps=return_comps,
        )

    def get_binned_emission_pix(
        self,
        freq: FrequencyOrWavelength,
        pixels: Pixels,
        nside: int,
        obs_time: Time,
        obs: str = "earth",
        obs_pos: u.Quantity[u.AU] | None = None,
        weights: Sequence[float] | npt.NDArray[np.floating] | None = None,
        return_comps: bool = False,
        coord_in: Literal["E", "G", "C"] = "E",
    ) -> u.Quantity[u.MJy / u.sr]:
        """Return the simulated binned zodiacal Emission given pixel numbers.

        The pixel numbers represent the pixel indicies on a HEALPix grid with resolution
        given by `nside`. The emission is binned to a HEALPix map with resolution given by
        `nside`.

        Args:
            freq: Delta frequency/wavelength or a sequence of frequencies corresponding to
                a bandpass over which to evaluate the zodiacal emission. The frequencies
                must be strictly increasing.
            pixels: HEALPix pixel indicies representing points on the celestial sphere.
            nside: HEALPix resolution parameter of the pixels and the binned map.
            obs_time: Time of observation.
            obs: Name of the Solar System observer. A list of all support observers (for a
                given ephemeridis) is specified in `supported_observers` attribute of the
                `Zodipy` instance. Defaults to 'earth'.
            obs_pos: The heliocentric ecliptic cartesian position of the observer in AU.
                Overrides the `obs` argument. Default is None.
            weights: Bandpass weights corresponding the the frequencies in `freq`. The weights
                are assumed to be given in spectral radiance units (Jy/sr).
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            coord_in: Coordinate frame of the input pointing. Assumes 'E' (ecliptic
                coordinates) by default.

        Returns:
            emission: Simulated zodiacal emission in units of 'MJy/sr'.

        """
        pixels = get_validated_pix(pixels=pixels, nside=nside)

        unique_pixels, counts = np.unique(pixels, return_counts=True)
        unit_vectors = get_unit_vectors_from_pixels(
            coord_in=coord_in,
            pixels=unique_pixels,
            nside=nside,
        )

        return self._compute_emission(
            freq=freq,
            weights=weights,
            obs=obs,
            obs_time=obs_time,
            obs_pos=obs_pos,
            unit_vectors=unit_vectors,
            indicies=counts,
            binned=True,
            pixels=unique_pixels,
            nside=nside,
            return_comps=return_comps,
        )

    def _compute_emission(
        self,
        freq: FrequencyOrWavelength,
        weights: Sequence[float] | npt.NDArray[np.floating] | None,
        obs: str,
        obs_time: Time,
        unit_vectors: npt.NDArray[np.float64],
        indicies: npt.NDArray[np.int64],
        binned: bool = False,
        obs_pos: u.Quantity[u.AU] | None = None,
        pixels: npt.NDArray[np.int64] | None = None,
        nside: int | None = None,
        return_comps: bool = False,
    ) -> u.Quantity[u.MJy / u.sr]:
        """Compute the component-wise zodiacal emission."""
        bandpass = validate_and_get_bandpass(
            freq=freq,
            weights=weights,
            model=self._ipd_model,
            extrapolate=self.extrapolate,
        )

        # Get model parameters, some of which have been interpolated to the given
        # frequency or bandpass.
        source_parameters = SOURCE_PARAMS_MAPPING[type(self._ipd_model)](
            bandpass, self._ipd_model, self._interpolator
        )

        observer_position, earth_position = get_obs_and_earth_positions(
            obs=obs, obs_time=obs_time, obs_pos=obs_pos
        )

        # Get the integration limits for each zodiacal component (which may be
        # different or the same depending on the model) along all line of sights.
        start, stop = get_line_of_sight_start_and_stop_distances(
            components=self._ipd_model.comps.keys(),
            unit_vectors=unit_vectors,
            obs_pos=observer_position,
        )

        density_partials = construct_density_partials_comps(
            comps=self._ipd_model.comps,
            dynamic_params={"X_earth": earth_position},
        )

        # Make table of pre-computed bandpass integrated blackbody emission.
        bandpass_interpolatation_table = get_bandpass_interpolation_table(bandpass)

        common_integrand = partial(
            EMISSION_MAPPING[type(self._ipd_model)],
            X_obs=observer_position,
            bp_interpolation_table=bandpass_interpolatation_table,
            **source_parameters["common"],
        )

        if self.parallel:
            n_proc = multiprocessing.cpu_count() if self.n_proc is None else self.n_proc

            unit_vector_chunks = np.array_split(unit_vectors, n_proc, axis=-1)
            integrated_comp_emission = np.zeros((len(self._ipd_model.comps), unit_vectors.shape[1]))
            with multiprocessing.get_context(SYS_PROC_START_METHOD).Pool(processes=n_proc) as pool:
                for idx, comp_label in enumerate(self._ipd_model.comps.keys()):
                    stop_chunks = np.array_split(stop[comp_label], n_proc, axis=-1)
                    if start[comp_label].size == 1:
                        start_chunks = [start[comp_label]] * n_proc
                    else:
                        start_chunks = np.array_split(start[comp_label], n_proc, axis=-1)
                    comp_integrands = [
                        partial(
                            common_integrand,
                            u_los=np.expand_dims(unit_vectors, axis=-1),
                            start=np.expand_dims(start, axis=-1),
                            stop=np.expand_dims(stop, axis=-1),
                            get_density_function=density_partials[comp_label],
                            **source_parameters[comp_label],
                        )
                        for unit_vectors, start, stop in zip(
                            unit_vector_chunks, start_chunks, stop_chunks
                        )
                    ]

                    proc_chunks = [
                        pool.apply_async(
                            _integrate_gauss_quad,
                            args=(comp_integrand, *self._gauss_points_and_weights),
                        )
                        for comp_integrand in comp_integrands
                    ]

                    integrated_comp_emission[idx] += (
                        np.concatenate([result.get() for result in proc_chunks])
                        * 0.5
                        * (stop[comp_label] - start[comp_label])
                    )

        else:
            integrated_comp_emission = np.zeros((len(self._ipd_model.comps), unit_vectors.shape[1]))
            unit_vectors_expanded = np.expand_dims(unit_vectors, axis=-1)

            for idx, comp_label in enumerate(self._ipd_model.comps.keys()):
                comp_integrand = partial(
                    common_integrand,
                    u_los=unit_vectors_expanded,
                    start=np.expand_dims(start[comp_label], axis=-1),
                    stop=np.expand_dims(stop[comp_label], axis=-1),
                    get_density_function=density_partials[comp_label],
                    **source_parameters[comp_label],
                )

                integrated_comp_emission[idx] = (
                    _integrate_gauss_quad(comp_integrand, *self._gauss_points_and_weights)
                    * 0.5
                    * (stop[comp_label] - start[comp_label])
                )

        emission = np.zeros(
            (
                len(self._ipd_model.comps),
                hp.nside2npix(nside) if binned else indicies.size,
            )
        )
        if binned:
            emission[:, pixels] = integrated_comp_emission
        else:
            emission = integrated_comp_emission[:, indicies]

        if self.solar_cut is not None:
            ang_dist = hp.rotator.angdist(-observer_position.flatten(), unit_vectors)
            solar_mask = ang_dist < self.solar_cut.value
            if binned and pixels is not None:
                emission[:, pixels[solar_mask]] = self.solar_cut_fill_value
            else:
                emission[:, solar_mask[indicies]] = self.solar_cut_fill_value

        emission = (emission << SPECIFIC_INTENSITY_UNITS).to(u.MJy / u.sr)

        return emission if return_comps else emission.sum(axis=0)

    def __repr__(self) -> str:
        repr_str = f"{self.__class__.__name__}("
        for attribute_name, attribute in self.__dict__.items():
            if attribute_name.startswith("_"):
                continue
            repr_str += f"{attribute_name}={attribute!r}, "

        return repr_str[:-2] + ")"


def _integrate_gauss_quad(
    fn: Callable[[float], npt.NDArray[np.float64]],
    points: npt.NDArray[np.float64],
    weights: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Integrate a function using Gauss-Legendre quadrature."""
    return np.squeeze(sum(fn(x) * w for x, w in zip(points, weights)))



import astropy.constants as const
import astropy.units as u
import numpy as np

h: float = const.h.value
c: float = const.c.value
k_B: float = const.k_B.value

SPECIFIC_INTENSITY_UNITS = u.W / u.Hz / u.m**2 / u.sr

R_0 = np.finfo(np.float64).eps
R_MARS = 1.5237
R_VERITAS = 2.16
R_EOS = 3.02
R_THEMIS = 3.14
R_EARTH = 1
R_ASTEROID_BELT = 3.137
R_NARROW_BANDS = 3.015
R_JUPITER = 5.2
R_KUIPER_BELT = 30
DISTANCE_FROM_EARTH_TO_L2 = u.Quantity(0.009896235034000056, u.AU)

N_INTERPOLATION_POINTS = 100
MIN_INTERPOLATION_GRID_TEMPERATURE = 40
MAX_INTERPOLATION_GRID_TEMPERATURE = 550



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Sequence

import astropy.units as u
import numpy as np

from zodipy._constants import (
    MAX_INTERPOLATION_GRID_TEMPERATURE,
    MIN_INTERPOLATION_GRID_TEMPERATURE,
    N_INTERPOLATION_POINTS,
)
from zodipy._source_funcs import get_blackbody_emission
from zodipy._validators import get_validated_and_normalized_weights, get_validated_freq

if TYPE_CHECKING:
    import numpy.typing as npt

    from zodipy._ipd_model import InterplanetaryDustModel
    from zodipy._types import FrequencyOrWavelength


@dataclass
class Bandpass:
    frequencies: FrequencyOrWavelength
    weights: npt.NDArray[np.float64]

    def integrate(self, quantity: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """Integrate a quantity over the bandpass."""
        return np.trapz(self.weights * quantity, self.frequencies.value, axis=-1)

    def switch_convention(self) -> None:
        """Switch the bandpass from frequency to wavelength or the other way around."""
        self.frequencies = self.frequencies.to(
            u.micron if self.frequencies.unit.is_equivalent(u.Hz) else u.Hz,
            equivalencies=u.spectral(),
        )
        if self.frequencies.size > 1:
            self.frequencies = np.flip(self.frequencies)
            self.weights = np.flip(self.weights)
            self.weights /= np.trapz(self.weights, self.frequencies.value)


def validate_and_get_bandpass(
    freq: FrequencyOrWavelength,
    weights: Sequence[float] | npt.NDArray[np.floating] | None,
    model: InterplanetaryDustModel,
    extrapolate: bool,
) -> Bandpass:
    """Validate bandpass and return a `Bandpass`."""
    freq = get_validated_freq(freq, model, extrapolate)
    normalized_weights = get_validated_and_normalized_weights(weights, freq)

    return Bandpass(freq, normalized_weights)


def get_bandpass_interpolation_table(
    bandpass: Bandpass,
    n_points: int = N_INTERPOLATION_POINTS,
    min_temp: float = MIN_INTERPOLATION_GRID_TEMPERATURE,
    max_temp: float = MAX_INTERPOLATION_GRID_TEMPERATURE,
) -> npt.NDArray[np.float64]:
    """Pre-compute the bandpass integrated blackbody emission for a grid of temperatures."""
    if not bandpass.frequencies.unit.is_equivalent(u.Hz):
        bandpass.switch_convention()

    bandpass_integrals = np.zeros(n_points)
    temperature_grid = np.linspace(min_temp, max_temp, n_points)
    for idx, temp in enumerate(temperature_grid):
        blackbody_emission = get_blackbody_emission(bandpass.frequencies.value, temp)
        bandpass_integrals[idx] = (
            bandpass.integrate(blackbody_emission)
            if bandpass.frequencies.size > 1
            else blackbody_emission
        )

    return np.asarray([temperature_grid, bandpass_integrals])



from typing import Sequence, Union

import astropy.units as u
import numpy as np
import numpy.typing as npt

Pixels = Union[int, Sequence[int], npt.NDArray[np.integer]]
SkyAngles = Union[u.Quantity[u.deg], u.Quantity[u.rad]]
FrequencyOrWavelength = Union[u.Quantity[u.Hz], u.Quantity[u.m]]
ParameterDict = dict



from typing import Tuple, Union

import astropy.units as u
import numpy as np
import numpy.typing as npt
from astropy.coordinates import HeliocentricMeanEcliptic, get_body
from astropy.time import Time

from zodipy._constants import DISTANCE_FROM_EARTH_TO_L2


@u.quantity_input
def get_obs_and_earth_positions(
    obs: str, obs_time: Time, obs_pos: Union[u.Quantity[u.AU], None]
) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Return the position of the observer and the Earth (3, `n_pointing`, `n_gauss_quad_degree`).

    The lagrange point SEMB-L2 is not included in any of the current available
    ephemerides. We implement an approximation to its position, assuming that
    SEMB-L2 is at all times located at a fixed distance from Earth along the vector
    pointing to Earth from the Sun.
    """
    earth_position = _get_earth_position(obs_time)
    if obs_pos is None:
        obs_position = _get_observer_position(obs, obs_time, earth_position)
    else:
        obs_position = obs_pos.to(u.AU)

    return obs_position.reshape(3, 1, 1).value, earth_position.reshape(3, 1, 1).value


def _get_earth_position(obs_time: Time) -> u.Quantity[u.AU]:
    """Return the position of the Earth given an ephemeris and observation time."""
    earth_skycoordinate = get_body("earth", obs_time)
    earth_skycoordinate = earth_skycoordinate.transform_to(HeliocentricMeanEcliptic)
    return earth_skycoordinate.cartesian.xyz.to(u.AU)


def _get_observer_position(
    obs: str, obs_time: Time, earth_pos: u.Quantity[u.AU]
) -> u.Quantity[u.AU]:
    """Return the position of the Earth and the observer."""
    if obs.lower() == "semb-l2":
        return _get_sun_earth_moon_barycenter(earth_pos)

    observer_skycoordinate = get_body(obs, obs_time)
    observer_skycoordinate = observer_skycoordinate.transform_to(HeliocentricMeanEcliptic)

    return observer_skycoordinate.cartesian.xyz.to(u.AU)


def _get_sun_earth_moon_barycenter(
    earth_position: u.Quantity[u.AU],
) -> u.Quantity[u.AU]:
    """Return the approximated position of SEMB-L2 given Earth's position."""
    r_earth = np.linalg.norm(earth_position)
    earth_unit_vector = earth_position / r_earth
    semb_l2_length = r_earth + DISTANCE_FROM_EARTH_TO_L2

    return earth_unit_vector * semb_l2_length
